nextflow.enable.dsl = 2

/*
 * Channels
*/
Channel
  .fromPath( params.fastq1 )
  .set { fastq1 }
Channel
  .fromPath( params.fastq2 )
  .set { fastq2 }
Channel
  .fromPath( params.reference_fasta )
  .set{ reference_fasta }
Channel
    .fromPath( params.sample )
    .set{ sample }
Channel
    .fromPath( params.bed )
    .set{ bed }
Channel
    .fromPath( params.blacklist )
    .set{ blacklist }
Channel
    .fromPath( params.manta_filter_script )
    .set{ manta_filter_script }
Channel
    .fromPath( params.gridss_filter_script )
    .set{ gridss_filter_script }
Channel
    .fromPath( params.structural_variants_script )
    .set{ structural_variants_script }
Channel
    .fromPath( params.batch_parser_script )
    .set{ batch_parser_script }
Channel
    .fromPath( params.exomedepth_filter_script )
    .set{ exomedepth_filter_script }
Channel
    .fromPath( params.exomeDepth_script )
    .set{ exomeDepth_script }
Channel
    .fromPath( params.codex_script )
    .set{ codex_script }
Channel
    .fromPath( params.codex_filter_script )
    .set{ codex_filter_script }

Channel
    .fromPath( params.collapse_script )
    .set{ collapse_script }

Channel
    .fromPath( params.merge_all_script )
    .set{ merge_all_script }

Channel
    .of( CNV_MANTA.out, CNV_GRIDSS.out, CNV_EXOMEDEPTH.out, CNV_CODEX.out )
    .set{ final_filtering }
/*
 * Define the default parameters
*/
params.threads_fastqc = 2
params.threads_fastp = 12
params.threads_samtools = 8
params.threads_gridss = 8
params.cut_right = true
params.cut_right_window_size = 5
params.cut_right_mean_quality = 24
params.trim_tail1 = 1
params.length_required = 70
params.min_mapping_quality = 10
params.bits_set = 4
params.assemblyFilename = "gridss.assembly.bam"
params.algoType = bwt_algorithm
params.read_group = true

/*
 * Includes for trimmed_fastq
 */
include {
  FASTQC as FASTQC_RAW_DATA
} from "./modules/fastqc"
include {
  FASTP
} from "./modules/fastp"
include {
  FASTQC as FASTQC_PRE_PROCESSED
} from "./modules/fastqc"

/*
 * Includes for bwa_index
 */
include {
  BWA_INDEX
} from "./modules/bwa_index"
include {
  SAMTOOLS_FAIDX
} from "./modules/samtools_faidx"

/*
 * Include for bwa_mem 
*/
include {
    BWA_MEM as BWA_MEM_READS
} from "./modules/bwa_mem"

/*
 * Include for samtools_view_sam2bam 
*/
include {
    SAMTOOLS_VIEW_SAM2BAM as SAMTOOLS_VIEW_SAM2BAM_PAIRED
} from "./modules/samtools_view_sam2bam"
include {
    SAMTOOLS_VIEW_SAM2BAM as SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_1
} from "./modules/samtools_view_sam2bam"
include {
    SAMTOOLS_VIEW_SAM2BAM as SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_2
} from "./modules/samtools_view_sam2bam"

/*
 * Include for samtools_sort
*/
include {
    SAMTOOLS_SORT as SAMTOOLS_SORT_PAIRED
} from "./modules/samtools_sort"
include {
    SAMTOOLS_SORT as SAMTOOLS_SORT_UNPAIRED_R1
} from "./modules/samtools_sort"
include {
    SAMTOOLS_SORT as SAMTOOLS_SORT_UNPAIRED_R2
} from "./modules/samtools_sort"

/*
 * Include for samtools_merge
*/
include {
    SAMTOOLS_MERGE as SAMTOOLS_MERGE_TOOL
} from "./modules/samtools_merge"

/*
 * Include for samtools_index
*/
include {
    SAMTOOLS_INDEX as SAMTOOLS_INDEX_TOOL
} from "./modules/samtools_index"

/*
 * Include for picard_markduplicates
*/
include {
    PICARD_MARKDUPLICATES as PICARD_MARKDUPLICATES_TOOL
} from "./modules/picard_markduplicates"

/*
 * Include for bam_filtering
*/
include {
    SAMTOOLS_VIEW
} from "./modules/samtools_view"

/*
 * Include for cnv_manta
*/
include {
    MANTA
} from "./modules/manta"
include {
    SVTOOLS
} from "./modules/svtools"
include {
    MANTA_FILTER
} from "./modules/manta_filter"
include {
    BEDOPS_UNION as MANTA_UNION
} from "./modules/bedops_union"
include {
    COLLAPSE as MANTA_MERGE
} from "./modules/collapse"

/*
 * Include for cnv_gridss
*/
include {
    GRIDSS
}from "./modules/gridss"
include {
    STRUCTURAL_VARIANTS
}from "./modules/structural_variants"
include {
    GRIDSS_FILTER
}from "./modules/gridss_filter"
include {
    BEDOPS_UNION as GRIDSS_UNION
} from "./modules/bedops_union"
include {
    COLLAPSE as GRIDSS_MERGE
} from "./modules/collapse"

/*
 * Include for cnv_exomedepth
*/
include {
    BATCH_PARSER
}from "./modules/batch_parser"
include {
    EXOME_DEPTH
}from "./modules/exome_depth"
include {
    EXOMEDEPTH_FILTER
}from "./modules/exomedepth_filter"
include {
    BEDOPS_UNION as EXOMEDEPTH_UNION
} from "./modules/bedops_union"
include {
    COLLAPSE as EXOMEDEPTH_MERGE
} from "./modules/collapse"

/*
 * Include for cnv_codex
*/
include {
    CODEX
}from "./modules/codex"
include {
    EXOMEDEPTH_FILTER
}from "./modules/codex_filter"
include {
    BEDOPS_UNION as CODEX_UNION
} from "./modules/bedops_union"
include {
    COLLAPSE as CODEX_MERGE
} from "./modules/collapse"

/*
 * Include for final_filtering
*/
include {
    BEDOPS_UNION as UNION
} from "./modules/bedops_union"
include {
    MERGE_ALL as MERGE
} from "./modules/merge_all"

/*
* Subworkflow trimmed_fastq
* Quality Control (raw data), Raw Data trimming and Quality Control (pre-processed)
*/
workflow SUB_TRIMMED_FASTQ {
  main:
    FASTQC_RAW_DATA(fastq1,
                    fastq2,
                    params.threads_fastqc)

    FASTP(fastq1,
          fastq2,
          params.threads_fastp,
          params.cut_right_window_size,
          params.cut_right_mean_quality,
          params.trim_tail1,
          params.length_required)

    FASTQC_PRE_PROCESSED(FASTP.out.paired_fastq1,
                         FASTP.out.paired_fastq2,
                         params.threads_fastqc)

  emit:
    reads = FASTP.out.reads
    unpaired_R1 = FASTP.out.unpaired_fastq1
    unpaired_R2 = FASTP.out.unpaired_fastq2
    html_report = FASTP.out.html_report
    json_report = FASTP.out.json_report
}

/*
* Subworkflow bwa_index
*/
workflow SUB_BWA_INDEX {
  main:
    BWA_INDEX(reference_fasta)
    SAMTOOLS_FAIDX(reference_fasta, BWA_INDEX.out)

  emit:
    output = SAMTOOLS_FAIDX.out
}

/*
* Subworkflow bwa_mem
*/
workflow SUB_BWA_MEM {
  take:
    reads
    index

  main:
    BWA_MEM_READS(reads,index)
  
  emit:
    output_reads = BWA_MEM_READS.out

}

/*
* Subworkflow samtools_view_sam2bam
*/
workflow SUB_SAMTOOLS_VIEW_SAM2BAM {
  take:
    paired
    unpaired_R1
    unpaired_R2
    threads

  main:
    SAMTOOLS_VIEW_SAM2BAM_PAIRED(paired,threads)

    SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_1(unpaired_R1,threads)

    SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_2(unpaired_R2,threads)

  emit:
    output_paired = SAMTOOLS_VIEW_SAM2BAM_PAIRED.out
    output_unpairedR1 = SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_1.out
    output_unpairedR2 = SAMTOOLS_VIEW_SAM2BAM_UNPAIRED_2.out
}

/*
* Subworkflow samtools_sort
*/
workflow SUB_SAMTOOLS_SORT {
  take:
    paired
    unpaired_R1
    unpaired_R2
    threads

  main:
    SAMTOOLS_SORT_PAIRED(paired,threads)

    SAMTOOLS_SORT_UNPAIRED_R1(unpaired_R1,threads)

    SAMTOOLS_SORT_UNPAIRED_R2(unpaired_R2,threads)

  emit:
    output_paired = SAMTOOLS_SORT_PAIRED.out
    output_unpairedR1 = SAMTOOLS_SORT_UNPAIRED_R1.out
    output_unpairedR2 = SAMTOOLS_SORT_UNPAIRED_R2.out
}

/*
* Subworkflow samtools_merge
*/
workflow SUB_SAMTOOLS_MERGE {
  take:
    paired
    unpaired_R1
    unpaired_R2
    threads

  main:
    SAMTOOLS_MERGE_TOOL(paired,
                        unpaired_R1,
                        unpaired_R2,
                        threads
    )
  
  emit:
    output = SAMTOOLS_MERGE_TOOL.out
}

/*
* Subworkflow samtools_index
*/
workflow SUB_SAMTOOLS_INDEX {
  take:
    input

  main:
    SAMTOOLS_INDEX_TOOL(input)

  emit:
    output = SAMTOOLS_INDEX_TOOL.out
}

/*
* Subworkflow picard_markduplicates
*/
workflow PICARD_MARKDUPLICATES {
    take:
        input
    
    main:
        PICARD_MARKDUPLICATES_TOOL(input)
        SAMTOOLS_INDEX(PICARD_MARKDUPLICATES_TOOL.out.alignments)
    
    emit:
        alignments = SAMTOOLS_INDEX.out
        metrics = PICARD_MARKDUPLICATES_TOOL.out.metrics
}

/*
* Subworkflow bam_filtering 
*/
workflow BAM_FILTERING {
    take:
      input      

    main:
        SAMTOOLS_VIEW( input )
        SAMTOOLS_INDEX( SAMTOOLS_VIEW.out )      

    emit:
      output = SAMTOOLS_INDEX.out
}

/*
* Subworkflow cnv_manta
*/
workflow CNV_MANTA {
    take:
       bams
       samples
       reference_genome
       bed

    main:
        MANTA(bams,
              reference_genome,
              bed)
        SVTOOLS( MANTA.out )
        MANTA_FILTER( manta_filter_script, SVTOOLS.out, samples )
        MANTA_UNION( MANTA_FILTER.out )
        MANTA_MERGE( collapse_script, MANTA_UNION.out )

    emit:
        output = MANTA_MERGE.out
}

/*
* Subworkflow cnv_gridss
*/
workflow CNV_GRIDSS {
    take:
        bams
        samples
        reference_genome

    main:
        GRIDSS(bams,
               reference_genome,
               blacklist)
        STRUCTURAL_VARIANTS( GRIDSS.out, structural_variants_script )
        GRIDSS_FILTER( gridss_filter_script, STRUCTURAL_VARIANTS.out, samples )
        GRIDSS_UNION( GRIDSS_FILTER.out )
        GRIDSS_MERGE( collapse_script, GRIDSS_UNION.out )

    emit:
        output = GRIDSS_MERGE.out

}

/*
* Subworkflow cnv_exomedepth
*/
workflow CNV_EXOMEDEPTH {
    take:
        bams
        samples
        reference_genome
        
    main:
        BATCH_PARSER( bams, samples, batch_parser_script )
        EXOME_DEPTH( bams, BATCH_PARSER.out, reference_genome, exomeDepth_script )
        EXOMEDEPTH_FILTER( exomedepth_filter_script, EXOME_DEPTH.out, samples )
        EXOMEDEPTH_UNION( EXOMEDEPTH_FILTER.out )
        EXOMEDEPTH_MERGE( collapse_script, EXOMEDEPTH_UNION.out )
        
    emit:
        output = EXOMEDEPTH_MERGE.out
}

/*
* Subworkflow cnv_codex
*/
workflow CNV_CODEX {
    take:
        bams
        samples
        bed

    main:
        BATCH_PARSER( bams, samples, batch_parser_script )
        CODEX( bams, BATCH_PARSER.out, bed, codex_script)
        CODEX_FILTER( CODEX.out, samples, codex_filter_script )
        CODEX_UNION( CODEX_FILTER.out )
        CODEX_MERGE( collapse_script, CODEX_UNION.out )

    emit:
        output = CODEX_MERGE.out
}

/*
* Subworkflow final_filtering
*/
workflow FINAL_FILTERING {
    take:
        input
        
    main:
        UNION( final_filtering )
        MERGE( UNION.out, merge_all_script )

    emit:
        output = MERGE.out
}

workflow{
  SUB_TRIMMED_FASTQ()
  SUB_BWA_INDEX()
  SUB_BWA_MEM(
    SUB_TRIMMED_FASTQ.out.reads,
    SUB_BWA_INDEX.out.output
  )
  SUB_SAMTOOLS_VIEW_SAM2BAM(
    SUB_BWA_MEM.out.output_paired,
    SUB_BWA_MEM.out.output_unpairedR1,
    SUB_BWA_MEM.out.output_unpairedR2,
    params.threads_samtools
  )
  SUB_SAMTOOLS_SORT(
    SUB_SAMTOOLS_VIEW_SAM2BAM.out.output_paired,
    SUB_SAMTOOLS_VIEW_SAM2BAM.out.output_unpairedR1,
    SUB_SAMTOOLS_VIEW_SAM2BAM.out.output_unpairedR2,
    params.threads_samtools
  )
  SUB_SAMTOOLS_MERGE(
    SUB_SAMTOOLS_SORT.out.output_paired,
    SUB_SAMTOOLS_SORT.out.output_unpairedR1,
    SUB_SAMTOOLS_SORT.out.output_unpairedR2,
    params.threads_samtools
  )
  SUB_SAMTOOLS_INDEX(
    SUB_SAMTOOLS_MERGE.out.output
  )
  PICARD_MARKDUPLICATES( SAMTOOLS_INDEX.out.output )
  BAM_FILTERING( PICARD_MARKDUPLICATES.out.alignments )
  CNV_MANTA (
      BAM_FILTERING.out,
      sample,
      BWA_INDEX.out,
      bed
  )
  CNV_GRIDSS (
      BAM_FILTERING.out,
      samples,
      BWA_INDEX.out
  )
  CNV_EXOMEDEPTH (
      BAM_FILTERING.out,
      samples,
      BWA_INDEX.out
  )
  CNV_CODEX (
      BAM_FILTERING.out,
      samples,
      bed
  )
  FINAL_FILTERING ( final_filtering )
}